"""Tools for reading and writing config files.

Support for new file extensions may be added by registering more config schemas."""
import json
import os
import sys
from typing import (
    ClassVar,
    List,
    Optional,
)

import yaml
import marsh


meta_key: str = os.environ.get('MARSH_META_KEY', '_meta_')
"""This key is used to store any metadata in a config.

The metadata is typically dropped after variable interpolations
have been performed."""


class ConfigSchemaMeta(marsh.schema.core.SchemaMeta['ConfigSchema']):

    def __call__(
        cls,  # noqa: B902
        value,
        *args,
        **kwargs,
    ):
        if not all(os.path.splitext(value)):
            raise ValueError(
                'unable to select config schema, given '
                f'path misses name or extension: {value}',
            )
        if cls is not ConfigSchema:
            return type(object).__call__(cls, value, *args, **kwargs)
        return cls.registry.match(value).build(value, *args, **kwargs)


class ConfigSchema(
    marsh.schema.core.Schema,
    metaclass=ConfigSchemaMeta,
):

    registry: ClassVar[marsh.schema.core.SchemaRegistry['ConfigSchema']] = \
        marsh.schema.core.SchemaRegistry(error=marsh.errors.ConfigFileError)

    @classmethod
    def match(
        cls,
        value: str,
    ) -> bool:
        raise NotImplementedError

    def dump(
        self,
        element: marsh.element.ElementType,
    ) -> None:
        encoded = self.dumps(element)
        with open(self.value, 'w') as fp:
            fp.write(encoded)

    def load(
        self,
    ) -> marsh.element.ElementType:
        with open(self.value, 'r') as fp:
            encoded = fp.read()
        return self.loads(encoded)

    def dumps(
        self,
        element: marsh.element.ElementType,
    ) -> str:
        raise NotImplementedError

    def loads(
        self,
        encoded: str,
    ) -> marsh.element.ElementType:
        raise NotImplementedError


@ConfigSchema.registry.register
class JSONConfigIO(ConfigSchema):

    @classmethod
    def match(
        cls,
        value: str,
    ) -> bool:
        return get_ext(value).lower() == '.json'

    def dump(
        self,
        element: marsh.element.ElementType,
    ) -> None:
        with open(self.value, 'w') as fp:
            json.dump(element, fp)

    def load(
        self,
    ) -> marsh.element.ElementType:
        with open(self.value, 'r') as fp:
            return json.load(fp)

    def dumps(
        self,
        element: marsh.element.ElementType,
    ) -> str:
        return json.dumps(element)

    def loads(
        self,
        encoded: str,
    ) -> marsh.element.ElementType:
        return json.loads(encoded)


@ConfigSchema.registry.register
class YAMLConfigIO(ConfigSchema):

    @classmethod
    def match(
        cls,
        value: str,
    ) -> bool:
        return get_ext(value).lower() in ('.yaml', '.yml')

    def dump(
        self,
        element: marsh.element.ElementType,
    ) -> None:
        with open(self.value, 'w') as fp:
            yaml.safe_dump(element, fp)

    def load(
        self,
    ) -> marsh.element.ElementType:
        with open(self.value, 'r') as fp:
            return yaml.load(fp, Loader=yaml.SafeLoader)

    def dumps(
        self,
        element: marsh.element.ElementType,
    ) -> str:
        return yaml.safe_dump(element)

    def loads(
        self,
        encoded: str,
    ) -> marsh.element.ElementType:
        return yaml.load(encoded, Loader=yaml.SafeLoader)


def has_ext(
    path: str,
) -> bool:
    return bool(os.path.splitext(path)[-1])


def strip_ext(
    path: str,
) -> str:
    if has_ext(path):
        return os.path.splitext(path)[0]
    return path


def get_ext(
    path_or_ext: str,
) -> str:
    a, b = os.path.splitext(path_or_ext)
    return b or a


def find_config(
    target: str,
) -> str:
    if has_ext(target):
        if not os.path.exists(target):
            raise marsh.errors.ConfigFileError(
                f'could not find config {target} ',
            )
        return target
    dpath = os.path.dirname(target)
    if not os.path.isdir(dpath):
        raise marsh.errors.ConfigFileError(
            f'unable to browse directory "{dpath}" '
            f'for named config "{target}"',
        )
    name = os.path.basename(target)
    names = []
    paths = []
    for candidate_fname in os.listdir(dpath):
        path = os.path.join(dpath, candidate_fname)
        if not os.path.isfile(path):
            continue
        if not config_type_is_supported(candidate_fname):
            continue
        candidate_name = strip_ext(candidate_fname)
        if candidate_name == name:
            paths.append(path)
        else:
            names.append(candidate_name)
    if paths:
        if len(paths) > 1:
            raise marsh.errors.ConfigFileError(
                f'ambiguous config "{target}" was resolved '
                'to more than one file: {}'.format(
                    ', '.join(f'"{os.path.basename(path)}"' for path in paths),
                ),
            )
        return paths[0]
    err_msg = (
        'could not find any supported config files in '
        f'directory "{dpath}" with name "{name}"'
    )
    closest = marsh.utils.get_closest(name, names)
    if closest:
        err_msg += f', did you mean "{os.path.join(dpath, closest)}"'
    raise marsh.errors.ConfigFileError(err_msg)


@marsh.utils.cache(maxsize=32, safe=False)
def config_type_is_supported(
    path: str,
) -> bool:
    try:
        ConfigSchema(path)
        return True
    except Exception:
        return False


def tree(
    path: str,
) -> str:
    """Get a tree representation of a directory recursively
    searched for configs.

    Arguments:
        path: The path to a directory.

    Returns:
        A tree representation of all available configs in that directory.
    """
    lines = [f'{os.path.basename(path)}']
    subtrees: List[str] = []
    configs: List[str] = []
    for name in os.listdir(path):
        if os.path.isdir(os.path.join(path, name)):
            subtree = tree(os.path.join(path, name))
            if subtree.count('\n') > 1:
                subtrees.append(name)
        else:
            if config_type_is_supported(name):
                configs.append(name)
    for i, config in enumerate(configs):
        prefix = '├── '
        if not subtrees and i == len(configs) - 1:
            prefix = '└── '
        lines.append(f'{prefix}{config}')
    for i, subtree in enumerate(subtrees):
        for j, line in enumerate(subtree.split('\n')):
            if j:
                prefix = '│   '
                if i == len(subtrees) - 1:
                    prefix = '    '
            else:
                prefix = '├── '
                if i == len(subtrees) - 1:
                    prefix = '└── '
            lines.append(f'{prefix}{line}')
    return '\n'.join(lines)


def load(
    name: str,
    *names: str,
    root: Optional[str] = None,
    resolve: bool = True,
    keep_meta: bool = False,
    concatenate: bool = False,
) -> marsh.element.ElementType:
    """Load one or more config files.

    When more than one config is loaded their
    contents are merged and the values of the
    succeeding config takes priority for fields
    where a merge can not be performed.

    File extensions may be omitted.

    Arguments:
        name: Name of a config file.
        *names: Any additional config files.
        root: A root working directory for the
            config files. Defaults to the current
            working directory.
        resolve: Resolve omegaconf interpolations.
        keep_meta: Keep meta field if it exists.
        concatenate: Concatenate sequences when
            multiple configs are loaded and merged
            instead of replacing the previous value.

    Returns:
        The loaded config.
    """
    configs = []
    for name in (name,) + names:
        if root and not name.startswith('/'):
            name = os.path.join(root, name)
        path = find_config(name)
        configs.append(ConfigSchema(path).load())
    config = marsh.element.merge(*configs, concatenate=concatenate)
    if (
        resolve
        and (
            marsh.utils.is_sequence(config)
            or marsh.utils.is_mapping(config)
        )
    ):
        config = marsh.element.resolve(config)
    if not keep_meta:
        config = drop_meta(config)
    return config


def write(
    config: marsh.element.ElementType,
    path: str = '-',
) -> None:
    """Write a config to a file.

    By default the config is written to
    stdout in YAML format.

    Arguments:
        config: The config to write.
        path: A filepath to write to.
            If not specified, stdout is used.
    """
    if path == '-':
        yaml.dump(config, sys.stdout)
    else:
        ConfigSchema(path).dump(config)


def drop_meta(
    config: marsh.element.ElementType,
) -> marsh.element.ElementType:
    """Drop the meta key with all its values if present.

    Arguments:
        config: Config to drop meta data from.

    Returns:
        Input config without meta data.
    """
    if (
        marsh.utils.is_mapping(config)
        and meta_key in config
    ):
        config = dict(config)
        del config[meta_key]
    return config



import datetime
from typing import (
    Any,
    Optional,
    Union,
)

import dateparser

import marsh


@marsh.schema.register
class DatetimeMarshalSchema(marsh.schema.MarshalSchema):
    """Marshals into string formatted according to ISO."""

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return isinstance(
                value,
                (
                    datetime.time,
                    datetime.date,
                ),
            )
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return (
            ':class:`datetime.datetime`, '
            ':class:`datetime.time`, '
            ':class:`datetime.date`'
        )

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals a datetime object into formatted string '
            'according to ISO. See :func:`datetime.datetime.isoformat`, .'
            ':func:`datetime.time.isoformat` and :func:`datetime.date.isoformat`.'
        )

    def marshal(
        self,
    ) -> str:
        return self.value.isoformat()


@marsh.schema.register
class DatetimeUnmarshalSchema(
    marsh.schema.UnmarshalSchema[
        Union[datetime.date, datetime.time],
    ],
):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return issubclass(
                value,
                (
                    datetime.time,
                    datetime.date,
                ),
            )
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return (
            ':class:`datetime.datetime`, '
            ':class:`datetime.time`, '
            ':class:`datetime.date`'
        )

    @staticmethod
    def doc_static_description() -> str:
        return (
            'For string inputs the format must be compatible '
            'with :func:`dateparser.parse`. The date order is '
            'D M Y, not M D Y. A float value can also be '
            'used as input in which case it is treated as a unix '
            'timestamp. A mapping or sequence as input will be passed '
            'directly to datetime as keyword or positional arguments.'
        )

    def doc_type(
        self,
    ) -> str:
        return 'datetime'

    def doc_field_type(
        self,
    ) -> str:
        return '<datetime>'

    def doc_description(
        self,
    ) -> str:
        return self.doc_static_description()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> Union[datetime.time, datetime.date]:
        if marsh.utils.is_missing(element):
            raise marsh.errors.MissingValueError(
                type=self.value,
            )
        try:
            if marsh.utils.is_mapping(element):
                return self.value(**element)
            elif marsh.utils.is_sequence(element):
                return self.value(*element)
        except Exception as err:
            raise marsh.errors.UnmarshalError(
                f'failed to unmarshal: {err}',
                element=element,
                type=self.value,
            ) from err
        dt: Optional[datetime.datetime] = None
        try:
            ts = marsh.utils.cast_primitive(float, str(element))
            dt = datetime.datetime.fromtimestamp(
                ts,
                datetime.timezone.utc,
            )
        except ValueError:
            if isinstance(element, str):
                dt = dateparser.parse(element, settings={'DATE_ORDER': 'DMY'})
        if dt is None:
            raise marsh.errors.UnmarshalError(
                'failed to parse timestamp',
                element=element,
                type=self.value,
            )
        if self.value == datetime.time:
            return dt.time()
        # datetime is subclass of date
        return dt



import dataclasses
import re
import types
from typing import (
    Any,
    ForwardRef,
    Optional,
    Tuple,
    TypeVar,
    get_type_hints,
)

import marsh


_T = TypeVar('_T')


_FIELDS = getattr(dataclasses, '_FIELDS', '__dataclass_fields__')
_INITVAR_SENTINEL = dataclasses._FIELD_INITVAR  # type: ignore


def get_init_var_fields(
    dataclass: Any,
) -> Tuple[dataclasses.Field, ...]:
    return tuple(
        field
        for field
        in getattr(
            dataclass,
            _FIELDS,
        ).values()
        if field._field_type is _INITVAR_SENTINEL
    )


def is_generated_class_docstring(
    docstring: Optional[str],
) -> bool:
    if docstring:
        return bool(
            re.match(
                r'^[a-zA-Z0-9_]+\(([a-zA-Z0-9_]+: .+( = .+)?)*\)$',
                docstring,
            ),
        )
    return False


@marsh.schema.register
class DataclassMarshalSchema(marsh.schema.MarshalSchema):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return (
                dataclasses.is_dataclass(value)
                and marsh.utils.is_obj_instance(value)
            )
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':func:`~dataclasses.dataclass`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals a dataclass instance into a mapping '
            'with the field names as keys.'
        )

    def marshal(
        self,
    ) -> dict:
        return {
            field.name: marsh.marshal(getattr(self.value, field.name))
            for field in dataclasses.fields(self.value)
            if field.init
        }


@marsh.schema.register
class DataclassUnmarshalSchema(marsh.schema.template.StructuredUnmarshalSchema[_T]):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        try:
            annotations = get_type_hints(marsh.utils.get_type(self.value))
        except Exception:
            annotations = {}
        schemas = {}

        def resolve_type(
            type_: Any,
        ) -> Any:
            try:
                if isinstance(type_, (str, ForwardRef)):
                    type_ = annotations.get(field.name, type_)
            except Exception:
                pass
            type_is_str = False
            try:
                type_is_str = isinstance(type_, str)
            except Exception:
                pass
            if type_is_str:
                raise marsh.errors.MarshError(
                    f'failed to resolve type from string: "{type_}"',
                )
            type_is_initvar = False
            try:
                type_is_initvar = isinstance(type_, dataclasses.InitVar)
            except Exception:
                pass
            if type_is_initvar:
                return resolve_type(type_.type)
            return type_

        for field in (
            dataclasses.fields(self.value)
            + get_init_var_fields(self.value)
        ):
            if not field.init:
                continue
            with marsh.errors.prepend(field.name):
                type_ = resolve_type(field.type)
                schemas[field.name] = marsh.schema.UnmarshalSchema(
                    type_,
                    default=(
                        marsh.MISSING
                        if field.default == dataclasses.MISSING
                        else field.default
                    ),
                    default_factory=(
                        marsh.MISSING
                        if field.default_factory == dataclasses.MISSING
                        else field.default_factory
                    ),
                )
        self.schemas = types.MappingProxyType(schemas)

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return (
                dataclasses.is_dataclass(value)
                and not marsh.utils.is_obj_instance(value)
            )
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':func:`~dataclasses.dataclass`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Expects a mapping input with the dataclass '
            'field names as keys.'
        )

    def doc_description(
        self,
    ) -> Optional[str]:
        doc = getattr(self.value, '__doc__', None)
        if not doc or is_generated_class_docstring(doc):
            return None
        return doc



from typing import Any

import marsh


@marsh.schema.register
class BytesMarshalSchema(marsh.schema.MarshalSchema):
    """Marshals into URL-safe base64 encodings from bytes."""

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return isinstance(value, bytes)
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`bytes`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals bytes into a URL-safe '
            'base64-encoded string.'
        )

    def marshal(
        self,
    ) -> str:
        return marsh.utils.bytes_to_base64(self.value)


@marsh.schema.register
class BytesUnmarshalSchema(marsh.schema.UnmarshalSchema[bytes]):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return issubclass(value, bytes)
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`bytes`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'A valid (optionally URL-safe) base64-encoded '
            'string is expected as input.'
        )

    def doc_type(
        self,
    ) -> str:
        return 'bytes'

    def doc_field_type(
        self,
    ) -> str:
        return '<bytes>'

    def doc_description(
        self,
    ) -> str:
        return self.doc_static_description()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> bytes:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            raise marsh.errors.MissingValueError(
                type=self.value,
            )
        if not marsh.utils.is_primitive(element):
            raise marsh.errors.UnmarshalError(
                (
                    f'expected a primitive value, got: '
                    f'{marsh.utils.get_type_name(element)}'
                ),
                element=element,
                type=self.value,
            )
        try:
            return marsh.utils.base64_to_bytes(str(element))
        except Exception:
            raise marsh.errors.UnmarshalError(
                'failed to unmarshal bytes',
                element=element,
                type=self.value,
            )



from typing import (
    Any,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)


import marsh


@marsh.schema.register
class TypeVarUnmarshalSchema(marsh.schema.UnmarshalSchema[Any]):

    value: TypeVar  # type: ignore
    schema: marsh.schema.UnmarshalSchema

    def __init__(
        self,
        value: Any,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(value, *args, **kwargs)
        self.schema = marsh.schema.UnmarshalSchema(
            self.resolve_type(),
            *args,
            **kwargs,
        )

    def __str__(
        self,
    ) -> str:
        return str(self.schema)

    def __repr__(
        self,
    ) -> str:
        return repr(self.schema)

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return bool(
                isinstance(value, TypeVar)
                and (
                    value.__bound__ is not None
                    or value.__constraints__
                ),
            )
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~typing.TypeVar`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unwraps the :class:`~typing.TypeVar`, using its bound '
            'or constraints when unmarshaling. If '
            'no bound or constraints are present '
            'the type can not be unmarshaled.'
        )

    def resolve_type(
        self,
    ) -> Any:
        if self.value.__constraints__:
            if len(self.value.__constraints__) == 1:
                return self.value.__constraints__[0]
            return Union[self.value.__constraints__]  # pyright: ignore
        return self.value.__bound__

    def select(
        self,
        path: str,
    ) -> marsh.schema.UnmarshalSchema:
        return self.schema.select(path)

    def doc(
        self,
        depth: int = 0,
    ) -> marsh.schema.UnmarshalSchema.Doc:
        return self.schema.doc(depth)

    def doc_type(
        self,
    ) -> str:
        return self.schema.doc_type()

    def doc_field_type(
        self,
    ) -> str:
        return self.schema.doc_field_type()

    def doc_default(
        self,
    ) -> Optional[str]:
        return self.schema.doc_default()

    def doc_description(
        self,
    ) -> Optional[str]:
        return self.schema.doc_description()

    def doc_fields(
        self,
        depth: int,
    ) -> Optional[Mapping[str, marsh.schema.UnmarshalSchema.Doc.Field]]:
        return self.schema.doc_fields(depth)

    def doc_special_fields(
        self,
    ) -> Optional[Sequence[marsh.schema.UnmarshalSchema.Doc.SpecialField]]:
        return self.schema.doc_special_fields()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> Any:
        return self.schema.unmarshal(element)



from typing import (
    Any,
    Iterable,
    Sequence,
    Type,
    Union,
    cast,
)

import marsh


UnmarshalDocType = Union[
    Type[Any],
    marsh.schema.UnmarshalSchema,
    marsh.schema.UnmarshalSchema.Doc,
]


RegisteredSchemasType = Union[
    Type[marsh.schema.core.Schema],
    marsh.schema.core.SchemaRegistry,
    Iterable[Type[marsh.schema.core.Schema]],
]


def get_unmarshal_doc(
    source: UnmarshalDocType,
    depth: int,
) -> marsh.schema.UnmarshalSchema.Doc:
    try:
        if isinstance(source, marsh.schema.UnmarshalSchema.Doc):
            return source
    except TypeError:
        pass
    try:
        if isinstance(source, marsh.schema.UnmarshalSchema):
            return source.doc(depth)
    except TypeError:
        pass
    return marsh.schema.UnmarshalSchema(source).doc(depth)


def get_registered_schemas(
    source: RegisteredSchemasType,
) -> Sequence[Type[marsh.schema.core.Schema]]:
    try:
        if issubclass(source, marsh.schema.core.Schema):  # type: ignore
            return tuple(
                cast(
                    Type[marsh.schema.core.Schema],
                    source,
                ).registry,
            )
    except TypeError:
        pass
    try:
        if isinstance(source, marsh.schema.core.SchemaRegistry):
            return tuple(source)
    except TypeError:
        pass
    return tuple(
        cast(
            Iterable[Type[marsh.schema.core.Schema]],
            source,
        ),
    )



from typing import (
    Optional,
    Sequence,
    TypeVar,
)

import marsh
from .. import UnmarshalSchema


_T = TypeVar('_T')


class UnionUnmarshalSchema(UnmarshalSchema[_T]):
    """A schema for a union of types.

    The order of the types is reflected in the
    order of attempted unmarshals. The value of
    the first type that is successfully unmarshaled
    is returned.

    If all types failed to unmarshal an error is raised."""

    schemas: Sequence[UnmarshalSchema]
    """The schemas for the union of types."""

    @property
    def optional_schema(
        self,
    ) -> Optional[UnmarshalSchema]:
        if (
            len(self.schemas) == 2
            and marsh.utils.is_optional(self.value)
        ):
            for schema in self.schemas:
                if schema.value not in (None, type(None)):
                    return schema
        return None

    def __str__(
        self,
    ) -> str:
        return ' | '.join(map(str, self.schemas))

    def __repr__(
        self,
    ) -> str:
        return f'{super().__repr__()[:-1]}, schemas={tuple(map(repr, self.schemas))})'

    def doc_field_type(
        self,
    ) -> str:
        return ' | '.join(schema.doc_field_type() for schema in self.schemas)

    def select(
        self,
        path: str,
    ) -> UnmarshalSchema:
        schema = self.optional_schema
        if schema is not None:
            return schema.select(path)
        if not path:
            return self
        raise marsh.errors.PathError(
            f'{marsh.utils.get_type_name(self.value)}: ambiguous '
            'traversal route (can not traverse union nodes)',
        )

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _T:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
        schema = self.optional_schema
        if schema is not None:
            try:
                return schema.unmarshal(element)
            except Exception:
                if (
                    element is None
                    or marsh.utils.is_missing(element)
                ):
                    return None  # type: ignore
                raise
        for schema in self.schemas:
            try:
                return schema.unmarshal(element)
            except Exception:
                pass
        if marsh.utils.is_missing(element):
            raise marsh.errors.MissingValueError(
                type=self.value,
            )
        raise marsh.errors.UnmarshalError(
            'failed to unmarshal element',
            element=element,
            type=self.value,
        )



import numbers
from typing import (
    Any,
    Literal,
    Mapping,
    Tuple,
    Union,
)

import marsh


Arg = Union[
    float,
    Tuple[()],
    Tuple[float],
    Tuple[float, float],
    Mapping[Literal['real', 'imag'], float],
]


@marsh.schema.register
class ComplexMarshalSchema(marsh.schema.MarshalSchema):

    value: numbers.Complex

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return isinstance(value, numbers.Complex)
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`complex`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals a complex value into a dictionary with keys '
            '``"real"`` and ``"imag"`` containing float values.'
        )

    def marshal(
        self,
    ) -> Union[float, dict]:
        c = complex(self.value)
        if c.imag == 0:
            return c.real
        return dict(
            real=c.real,
            imag=c.imag,
        )


@marsh.schema.register
class ComplexUnmarshalSchema(marsh.schema.UnmarshalSchema[complex]):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            # let the generic unmarshal schema for
            # callables handle any subclasses of
            # numbers.Complex as they may contain
            # more arguments than just `real` and `imag`.
            return (
                issubclass(value, complex)
                or value == numbers.Complex
            )
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`complex`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Expects a mapping with optional keys "real" '
            'and/or "imag" with float values or a sequence containing '
            'up to two float values, the first being the real value '
            'and the second being the imaginary value. If passed '
            'a float value instead of a mapping or sequence, it is '
            'treated as the real value.'
        )

    def doc_field_type(
        self,
    ) -> str:
        return '<complex>'

    def doc_description(
        self,
    ) -> str:
        return self.doc_static_description()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> complex:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            raise marsh.errors.MissingValueError(
                type=self.value,
            )
        try:
            arg: Arg = marsh.unmarshal(Arg, element)  # type: ignore
        except Exception:
            raise marsh.errors.UnmarshalError(
                (
                    'failed to unmarshal input element '
                    'into a float, tuple of floats or mapping '
                    'with optional keys `name` and `level`'
                ),
                element=element,
                type=self.value,
            )
        if isinstance(arg, float):
            return complex(arg)
        if marsh.utils.is_sequence(arg):
            return complex(*arg)
        return complex(**arg)  # type: ignore



from . import common


def format_types(
    source: common.RegisteredSchemasType,
    heading: int = 3,
) -> str:
    """Return a markdown-formatted string containing
    registered schema types that have static documention.

    A description below each type is included when available.

    Arguments:
        source: The source of the schemas.
        heading: The heading number used for each types.

    Returns:
        The formatted markdown.
    """
    assert heading >= 1, '`heading` must be a positive integer'
    sections = []
    type_prefix = heading * '#' + ' '
    for schema in common.get_registered_schemas(source):
        type_ = schema.doc_static_type()
        if not type_:
            continue
        sections.append(
            f'{type_prefix}{type_}\n'
            f'{schema.doc_static_description() or ""}',
        )
    return '\n\n'.join(sections)



import enum
from typing import (
    Any,
    TypeVar,
)

import marsh
from . import sequence


_E = TypeVar('_E', bound=enum.Enum)


@marsh.schema.register(lower_priority=sequence.SequenceUnmarshalSchema)
class EnumUnmarshalSchema(marsh.schema.template.UnionUnmarshalSchema[_E]):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.schemas = tuple(map(EnumValueUnmarshalSchema, self.value))

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return issubclass(value, enum.Enum)
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~enum.Enum`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Input must match one of the '
            'enum values (the name or value).'
        )


@marsh.schema.register
class EnumValueMarshalSchema(marsh.schema.MarshalSchema):

    value: enum.Enum

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return isinstance(value, enum.Enum)
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~enum.Enum`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals an enum instance into its name.'
        )

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        return self.value.name


@marsh.schema.register
class EnumValueUnmarshalSchema(marsh.schema.UnmarshalSchema[_E]):

    value: _E  # type: ignore

    def __str__(
        self,
    ) -> str:
        return self.value.name

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return isinstance(value, enum.Enum)
        except Exception:
            return False

    def doc_type(
        self,
    ) -> str:
        return marsh.utils.get_type_name(self.value)

    def doc_field_type(
        self,
    ) -> str:
        return f'"{self.value.name}"'

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _E:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            raise marsh.errors.MissingValueError(
                type=self.value,
            )
        if not marsh.utils.is_primitive(element):
            raise marsh.errors.UnmarshalError(
                (
                    'expected a primitive value for an enum, got: '
                ),
                element=element,
                type=self.value,
            )
        try:
            marsh.utils.cast_literal(self.value.name, element)
            return self.value
        except ValueError:
            pass
        try:
            marsh.utils.cast_literal(self.value.value, element)
            return self.value
        except ValueError:
            pass
        raise marsh.errors.UnmarshalError(
            (
                f'could not convert to enum name ({self.value.name}) '
                f'or enum value ({self.value.value})'
            ),
            element=element,
            type=self.value,
        )



import collections
import dataclasses
import importlib
import threading
from typing import (
    Any,
    Callable,
    DefaultDict,
    ForwardRef,
    Generic,
    Iterable,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    get_origin,
    overload,
)
from typing_extensions import TypeAlias

import marsh
from . import base


_T = TypeVar('_T')
_RelativePriority: TypeAlias = \
    marsh.utils.PriorityOrder.RelativePriority[Type['UnmarshalSchema']]


skip_cache_types = {
    int,
    float,
    str,
    bool,
    dict,
    list,
    tuple,
    None,
    set,
}


def resolve_forward_ref(
    ref: ForwardRef,
    cache: Iterable[Any],
) -> Any:
    name = ref.__forward_arg__
    mod = ''
    if '.' in name:
        mod, _, name = name.rpartition('.')
    if mod:
        try:
            return ref._evaluate(
                vars(importlib.import_module(mod)),
                None,
            )
        except Exception:
            pass
    for type_ in cache:
        if marsh.utils.get_type_name(type_) == name:
            return type_
        if hasattr(type_, '__module__'):
            try:
                return ref._evaluate(
                    vars(
                        importlib.import_module(
                            type_.__module__ + (f'.{mod}' if mod else ''),
                        ),
                    ),
                    None,
                )
            except Exception:
                pass
    raise marsh.errors.MarshError(f'could not resolve forward reference to a type: {ref}')


def resolve_type(
    value: Any,
    cache: Iterable[Any],
) -> Any:
    """Attempt to resolve unknown types (forward references).

    Returns the input type if it is not a forward reference.

    Arguments:
        value: The type to resolve.
        cache: A cache of types.

    Returns:
        The resolved type.
    """
    try:
        if isinstance(value, ForwardRef):
            value = resolve_forward_ref(
                value,
                cache,
            )
            return resolve_type(value, cache)
    except marsh.errors.MarshError:
        raise
    except Exception:
        pass
    return value


class UnmarshalSchemaRegistry(base.SchemaRegistry['UnmarshalSchema']):

    def cache_clear(
        self,
    ) -> None:
        self.match.cache_clear()  # type: ignore

    def cache_info(
        self,
    ) -> marsh.utils.CacheInfo:
        return self.match.cache_info,  # type: ignore

    @overload
    def register(
        self,
        schema: Type['UnmarshalSchema'],
        *,
        priority: int = 0,
        lower_priority: _RelativePriority = None,
        higher_priority: _RelativePriority = None,
        replace: bool = False,
    ) -> Type['UnmarshalSchema']:
        ...

    @overload
    def register(
        self,
        *,
        priority: int = 0,
        lower_priority: _RelativePriority = None,
        higher_priority: _RelativePriority = None,
        replace: bool = False,
    ) -> Callable[[Type['UnmarshalSchema']], Type['UnmarshalSchema']]:
        ...

    def register(
        self,
        schema: Optional[Type['UnmarshalSchema']] = None,
        *,
        priority: int = 0,
        lower_priority: _RelativePriority = None,
        higher_priority: _RelativePriority = None,
        replace: bool = False,
    ) -> Union[
        Callable[
            [Type['UnmarshalSchema']],
            Type['UnmarshalSchema'],
        ],
        Type['UnmarshalSchema'],
    ]:
        def wrapper(_schema):
            self.cache_clear()
            return self.priority.add(
                _schema,
                priority=priority,
                lower_priority=lower_priority,
                higher_priority=higher_priority,
                replace=replace,
            )
        if schema is None:
            return wrapper
        return wrapper(schema)

    @base.caches.new_callable_cache(
        name='marsh.schema.UnmarshalSchema.registry.match',
        typed=False,
        safe=True,
        binding='ignore',
    )
    def match(
        self,
        value: Any,
    ) -> base.SchemaSelection['UnmarshalSchema']:
        return super().match(value)


class UnmarshalSchemaMeta(base.SchemaMeta['UnmarshalSchema']):

    class CacheInfo(NamedTuple):
        registry: marsh.utils.CacheInfo
        recursion: marsh.utils.CacheInfo
        build: marsh.utils.CacheInfo

    _recursive_cache: DefaultDict[
        int,
        marsh.utils.ValueCache[
            Type[Any],
            base.SchemaSelection,
        ],
    ] = collections.defaultdict(marsh.utils.ValueCache)

    type_cache: marsh.utils.WeakTypeCache = \
        marsh.utils.WeakTypeCache(skip_types=skip_cache_types)

    registry: UnmarshalSchemaRegistry = UnmarshalSchemaRegistry(
        error=marsh.errors.UnmarshalError,
    )

    def __init_subclass__(
        metacls,  # noqa: B902
    ) -> None:
        metacls._recursive_cache = \
            collections.defaultdict(marsh.utils.ValueCache)
        metacls.type_cache = \
            marsh.utils.WeakTypeCache(skip_types=skip_cache_types)
        if (
            hasattr(metacls, 'registry')
            and metacls.registry is not UnmarshalSchema.registry
        ):
            return
        super().__init_subclass__()
        metacls.registry = UnmarshalSchemaRegistry()

    def __call__(
        cls,  # noqa: B902
        value,
        *args,
        **kwargs,
    ):
        if cls is not UnmarshalSchema:
            return type(object).__call__(cls, value, *args, **kwargs)
        metacls: 'UnmarshalSchemaMeta' = type(cls)
        value = resolve_type(value, cache=metacls.type_cache)
        if get_origin(value) is Union:
            return cls._build(
                value=value,
                *args,
                **kwargs,
            )
        return cls._cached_build(
            value=value,
            *args,
            **kwargs,
        )

    def _build(
        cls,  # noqa: B902
        value,
        *args,
        **kwargs,
    ):
        metacls: 'UnmarshalSchemaMeta' = type(cls)
        try:
            metacls.type_cache.add(value)
        except TypeError:
            pass
        recursive_cache = cls._recursive_cache[threading.get_ident()]
        if recursive_cache:
            try:
                return recursive_cache[value].build(value, *args, **kwargs)
            except KeyError:
                pass
        result = cls.registry.match(value)
        result.insert(0, RecursiveReferenceUnmarshalSchema)
        recursive_cache[value] = result
        return result[1:].build(value, *args, **kwargs)

    @base.caches.new_callable_cache(
        name='marsh.schema.UnmarshalSchema.__new__',
        safe=True,
        binding='ignore',
    )
    def _cached_build(
        cls,  # noqa: B902
        value,
        *args,
        **kwargs,
    ):
        return cls._build(
            value=value,
            *args,
            **kwargs,
        )

    def cache_clear(
        cls,  # noqa: B902
    ) -> None:
        cls.registry.cache_clear()
        cls._cached_build.cache_clear()  # type: ignore
        cls._recursive_cache.clear()
        cls.type_cache.clear()

    def cache_info(
        cls,  # noqa: B902
    ) -> 'UnmarshalSchemaMeta.CacheInfo':
        return dict(
            registry=cls.registry.cache_info(),
            recursion=cls._recursive_cache[threading.get_ident()].cache_info(),
            build=cls._cached_build.cache_info(),  # type: ignore
        )


class UnmarshalSchema(
    base.Schema,
    Generic[_T],
    metaclass=UnmarshalSchemaMeta,
):
    """Unmarshals JSON-like inputs into its value type.

    .. note::

        Static type checkers may give an error for
        abstract types
        (`issue <https://github.com/python/mypy/issues/5374>`_).

        Some type aliases are also not recognized as types.
        For the moment there is not better solution than
        using a simple ``# type: ignore`` comment.

    Arguments:
        value: The value to match/unmarshal.
        default: An immutable default value. Mutually
            exclusive with ``default_factory``.
        default_factory: Factory function for a mutable
            default value. Mutually exclusive with
            ``default``.
    """

    @dataclasses.dataclass
    class Doc:
        """Structured unmarshal documentation for a type."""

        @dataclasses.dataclass
        class Field:
            """Unmarshal documentation for a field/attribute
            of a type."""

            doc: 'UnmarshalSchema.Doc'
            """The field unmarshal documentation."""

            type: Optional[str] = None
            """Type string as should be displayed when this is
            a field of a parent."""

            description: Optional[str] = None
            """Description for field by parent."""

        @dataclasses.dataclass
        class SpecialField:
            """Unmarshal documentation for a special field.

            A special field may document any value.
            """

            value: str
            """The name of the value being documented."""

            description: Optional[str] = None
            """Its description."""

        type: Optional[str] = None
        """String representation of the type."""

        default: Optional[str] = None
        """Default value."""

        description: Optional[str] = None
        """Description for type."""

        fields: Optional[Mapping[str, 'UnmarshalSchema.Doc.Field']] = None
        """Named internal fields of the documented type."""

        special_fields: Optional[Sequence['UnmarshalSchema.Doc.SpecialField']] = None
        """Any other fields that do not follow the name/type/default
        style."""

    value: Type[_T]
    default: _T
    default_factory: Callable[..., _T]

    def __init__(
        self,
        *args,
        default: _T = marsh.MISSING,
        default_factory: Callable[..., _T] = marsh.MISSING,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.default = default
        self.default_factory = default_factory

    def __str__(
        self,
    ) -> str:
        return f'<{self.doc_type()}>'

    def __repr__(
        self,
    ) -> str:
        return (
            f'{self.__class__.__name__}(value={self.value}, '
            f'default={self.default}, default_factory={self.default_factory})'
        )

    def has_default(
        self,
    ) -> bool:
        """Evaluates if this schema holds a default value.

        Returns:
            ``True`` if there is a default value, else ``False``.
        """
        return not (
            marsh.utils.is_missing(self.default)
            and marsh.utils.is_missing(self.default_factory)
        )

    def get_default(
        self,
    ) -> _T:
        """Get the default value of this schema.

        If default value exists `marsh.MISSING`
        is returned instead.

        Returns:
            The default value or `marsh.MISSING`.
        """
        if not marsh.utils.is_missing(self.default_factory):
            return self.default_factory()
        return self.default

    def select(
        self,
        path: str,
    ) -> 'UnmarshalSchema':
        """Select a nested schema from this schema based on
        a dot-separated path. If not all of the path is consumed
        by this schema it is forwarded to the `select` method
        of the nested schema.

        Arguments:
            path: The path to traverse.

        Returns:
            The schema at the end of the path.
        """
        if not path:
            return self
        raise marsh.errors.PathError(
            f'{marsh.utils.get_type_name(self.value)}: '
            f'can not traverse the path "{path}"',
        )

    def doc(
        self,
        depth: int = 0,
    ) -> Doc:
        """Get the documentation for the type unmarshaled
        by this schema.

        Arguments:
            depth: How deep should documentation be fetched
                in nested schemas. If ``0``, only the fields
                and info for the type of this schema is returned.

        Returns:
            The documentation.
        """
        return self.Doc(
            type=self.doc_type(),
            default=self.doc_default(),
            description=self.doc_description(),
            fields=self.doc_fields(max(0, depth)),
            special_fields=self.doc_special_fields(),
        )

    def doc_type(
        self,
    ) -> str:
        """The name of the type unmarshaled by this schema.

        Returns:
            The type name.
        """
        return marsh.utils.get_type_name(self.value)

    def doc_field_type(
        self,
    ) -> str:
        """Alternative name of the type unmarshaled by this schema.

        This corresponds to the type that should be display in parent
        documentation when this is one of its fields.

        Returns:
            Alternative type name.
        """
        return self.doc_type()

    def doc_default(
        self,
    ) -> Optional[str]:
        """Document default value.

        The default implementation of this method
        returns :data:`None` if no default value exists
        and ``'...'`` if the default value is mutable.
        If the default value is immutable, an
        attempt is first made to marshal the default
        value before returning a string of its marshaled
        state. If this fails, a string of the default
        is returned.

        Returns:
            String of default.
        """
        if not self.has_default():
            return None
        if marsh.utils.is_missing(self.default):
            return '...'
        try:
            return str(marsh.marshal(self.default))
        except Exception:
            return str(self.default)

    def doc_description(
        self,
    ) -> Optional[str]:
        """A description for the type unmarshaled by this schema.

        Returns:
            The description.
        """
        if marsh.utils.is_typing_alias(self.value):
            return None
        try:
            if self.value in (int, float, bool, str, bytes, tuple, list, dict):
                return None
        except Exception:
            pass
        return marsh.utils.get_description(self.value)

    def doc_fields(
        self,
        depth: int,
    ) -> Optional[Mapping[str, Doc.Field]]:
        """Documentation for the fields of the type unmarshaled
        by this schema.

        No documentation is returned if no fields exist.

        Returns:
            Field documentation.
        """
        return None

    def doc_special_fields(
        self,
    ) -> Optional[Sequence[Doc.SpecialField]]:
        """Documentation for the special fields of the type unmarshaled
        by this schema.

        No documentation is returned if no special fields exist.

        Returns:
            Special field documentation.
        """
        return None

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _T:
        """Construct the object of this schema.

        Arguments:
            element: The input JSON-style data.

        Returns:
            The object initialized with the input data.
        """
        raise NotImplementedError


class WrapperUnmarshalSchema(
    UnmarshalSchema[_T],
    base.WrapperSchema[UnmarshalSchema[_T]],
):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        base.WrapperSchema.__init__(self, *args, **kwargs)

    def __str__(
        self,
    ) -> str:
        return str(self.schema)

    def __repr__(
        self,
    ) -> str:
        return repr(self.schema)

    def select(
        self,
        path: str,
    ) -> 'UnmarshalSchema':
        return self.schema.select(path)

    def doc(
        self,
        depth: int = 0,
    ) -> UnmarshalSchema.Doc:
        return self.schema.doc(depth)

    def doc_type(
        self,
    ) -> str:
        return self.schema.doc_type()

    def doc_field_type(
        self,
    ) -> str:
        return self.schema.doc_field_type()

    def doc_default(
        self,
    ) -> Optional[str]:
        return self.schema.doc_default()

    def doc_description(
        self,
    ) -> Optional[str]:
        return self.schema.doc_description()

    def doc_fields(
        self,
        depth: int,
    ) -> Optional[Mapping[str, UnmarshalSchema.Doc.Field]]:
        return self.schema.doc_fields(depth)

    def doc_special_fields(
        self,
    ) -> Optional[Sequence[UnmarshalSchema.Doc.SpecialField]]:
        return self.schema.doc_special_fields()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _T:
        return self.schema.unmarshal(element)


class RecursiveReferenceUnmarshalSchema(WrapperUnmarshalSchema):

    def __str__(
        self,
    ) -> str:
        return UnmarshalSchema.__str__(self)

    def __repr__(
        self,
    ) -> str:
        return UnmarshalSchema.__repr__(self)

    def doc(
        self,
        depth: int = 0,
    ) -> UnmarshalSchema.Doc:
        # Dont include fields in the documentation
        # of recursive types as that would repeat
        # the same documentation over when depth > 0
        return self.Doc(
            type=self.schema.doc_type(),
            default=self.schema.doc_default(),
            description=self.schema.doc_description(),
        )



"""The core functionality powering marshaling and unmarshaling."""
from .base import (
    Schema,
    SchemaMeta,
    SchemaRegistry,
    SchemaSelection,
)
from . import (
    marshal,
    unmarshal,
)


__all__ = (
    'Schema',
    'SchemaMeta',
    'SchemaRegistry',
    'SchemaSelection',
    'marshal',
    'unmarshal',
)



from typing import (
    Any,
    Union,
    cast,
)

import omegaconf

import marsh


@marsh.schema.register
class OmegaconfMarshalSchema(marsh.schema.MarshalSchema):

    value: Union[omegaconf.DictConfig, omegaconf.ListConfig]

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return isinstance(value, (omegaconf.DictConfig, omegaconf.ListConfig))
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~omegaconf.DictConfig`, :class:`~omegaconf.ListConfig`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals an omegaconf :class:`~omegaconf.DictConfig` '
            'or :class:`~omegaconf.ListConfig`, converting '
            'all enums in the container to strings.'
        )

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        return cast(
            marsh.element.ElementType,
            omegaconf.OmegaConf.to_container(
                self.value,
                enum_to_str=True,
            ),
        )



"""This module houses the mechanics that powers
marshaling and unmarshaling."""
from . import core
from .core.base import caches
from .core.marshal import MarshalSchema
from .core.unmarshal import UnmarshalSchema
from . import template
from . import namespace
from . import argument
from ._functional import register


__all__ = (
    'core',
    'caches',
    'MarshalSchema',
    'UnmarshalSchema',
    'template',
    'namespace',
    'argument',
    'register',
)



from typing import (
    Any,
    Mapping,
    NamedTuple,
    Sequence,
)

from . import template


class Arguments(NamedTuple):
    args: Sequence[Any]
    kwargs: Mapping[str, Any]


class ArgumentsUnmarshalSchema(
    template.CallableUnmarshalSchema[Arguments],
):

    def construct(
        self,
        *args,
        **kwargs,
    ) -> Arguments:
        return Arguments(
            args=args,
            kwargs=kwargs,
        )



"""Root of the marsh framework.

Exposes the basic functionality and modules of the framework."""
from omegaconf import MISSING
from . import errors
from . import utils
from . import element
from . import path
from . import schema
from . import config
from . import annotation
from ._functional import (
    marshal,
    unmarshal,
    unmarshal_args,
    cast,
    cast_args,
    main,
)

# import module containing baseline type schemas
# which are then registered and available in the framework
# for marshaling and unmarshaling.
import marsh.schema.types  # noqa: F401


__version__ = '0.2.4'


# singleton class, create a reference for easy access
namespaces = schema.namespace.Namespaces()
"""Allows access to any registered namespace."""


__all__ = (
    'MISSING',
    'errors',
    'utils',
    'element',
    'path',
    'shema',
    'config',
    'marshal',
    'unmarshal',
    'unmarshal_args',
    'cast',
    'cast_args',
    'main',
    '__version__',
    'namespaces',
    'annotation',
)



import functools
from typing import (
    Any,
    Callable,
    Optional,
    Type,
    TypeVar,
    overload,
)

import marsh
from . import entry_point


_T = TypeVar('_T')
_R = TypeVar('_R')
_C = TypeVar('_C', bound=Callable)


def marshal(
    value: Any,
) -> marsh.element.ElementType:
    """Serialize a value to a JSON-like object.

    Arguments:
        value: The input value to serialize.

    Returns:
        The JSON representation of the object.
    """
    schema = marsh.schema.MarshalSchema(value)
    return schema.marshal()


@overload
def unmarshal(
    type_: Type[_T],
    element: marsh.element.ElementType = marsh.MISSING,
) -> _T:
    ...


@overload
def unmarshal(
    type_: Type[_T],
    element: marsh.element.MappingElementType = marsh.MISSING,
    **kwargs,
) -> _T:
    ...


def unmarshal(
    type_: Type[_T],
    element: marsh.element.ElementType = marsh.MISSING,
    **kwargs,
) -> _T:
    """Unmarshal a JSON-like object to some type.

    .. note::

        Static type checkers may give an error for
        abstract types
        (`issue <https://github.com/python/mypy/issues/5374>`_).

        Some type aliases are also not recognized as types.
        For the moment there is not better solution than
        using a simple ``# type: ignore`` comment.

    Arguments:
        type_: The type to cast to.
        element: The JSON-like object.

    Returns:
        The unmarshaled value.
    """
    schema: marsh.schema.UnmarshalSchema[_T] = \
        marsh.schema.UnmarshalSchema(type_)
    if kwargs:
        if marsh.utils.is_missing(element):
            element = marshal(kwargs)
        if not marsh.utils.is_mapping(element):
            raise TypeError(
                'additional keyword arguments may only '
                'be used when the input element is a '
                'mapping type.',
            )
        else:
            element = dict(element)
            element.update(marshal(kwargs))  # type: ignore
    return schema.unmarshal(element)


def unmarshal_args(
    callable_: Callable,
    element: marsh.element.ElementType = marsh.MISSING,
) -> marsh.schema.argument.Arguments:
    """Return the unmarshaled arguments of the given function.

    Arguments:
        callable_: The function whos arguments to unmarshal.
        element: The input to the unmarshaler.

    Returns:
        Tuple of positional arguments and keyword arguments.
    """
    schema = marsh.schema.argument.ArgumentsUnmarshalSchema(callable_)
    return schema.unmarshal(element)


def cast(
    type_: Type[_T],
    value: Any = marsh.MISSING,
) -> _T:
    """Cast a value to some type.
    The input value is serialized to
    a JSON-like object and then used
    for constructing an instance of
    the specified type.

    .. note::

        Static type checkers may give an error for
        abstract types
        (`issue <https://github.com/python/mypy/issues/5374>`_).

        Some type aliases are also not recognized as types.
        For the moment there is not better solution than
        using a simple ``# type: ignore`` comment.

    Arguments:
        type_: The type to cast to.
        value: The value to cast.

    Returns:
        The casted value.
    """
    element = marshal(value)
    return unmarshal(type_, element)


def cast_args(
    callable_: Callable,
    value: Any = marsh.MISSING,
) -> marsh.schema.argument.Arguments:
    """Return the casted arguments of the given function.

    The given input will be marshaled before being unmarshaled
    into the function arguments.

    Arguments:
        callable_: The function whos arguments to unmarshal.
        value: The input.

    Returns:
        Tuple of positional arguments and keyword arguments.
    """
    element = marshal(value)
    return unmarshal_args(
        callable_,
        element,
    )


@overload
def main(
    *,
    setup_logging: bool = False,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    help_depth: int = 1,
) -> Callable[[_C], _C]:
    ...


@overload
def main(
    fn: _C,
    /,
    *,
    setup_logging: bool = False,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    help_depth: int = 1,
) -> _C:
    ...


@overload
def main(
    *,
    config: Type[_T],
    setup_logging: bool = False,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    help_depth: int = 1,
) -> Callable[[Callable[[_T], _R]], Callable[[_T], _R]]:
    ...


@overload
def main(
    fn: Callable[[_T], _R],
    /,
    *,
    config: Type[_T],
    setup_logging: bool = False,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    help_depth: int = 1,
) -> Callable[[_T], _R]:
    ...


def main(
    fn=None,
    /,
    *,
    config: Optional[Any] = None,
    setup_logging: bool = False,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    help_depth: int = 2,
):
    """Decorates a function, allowing it to be called as an entry
    point for a python script being run from the command line.

    If a config class is supplied, the decorated function will be passed
    an instance of that class as a single positional argument.
    If no config class is supplied, the arguments for the decorated
    function are filled out and passed to the function.

    Arguments:
        fn: The callable function to decorate. When decorating,
            this argument should not be explicitly specified.
        config: An optional config class.
        setup_loggin: Add console line arguments for logging
            level and format. Initiallizes logging through
            :func:`logging.basicConfig`.
        prog: Optional program name. Defaults to the first
            value of :attr:`sys.argv`.
        description: Optional description shown in the help message.
            Defaults to the docstring of the decorated function (or
            the config class if passed to the decorator). Can
            be set to :data:`None` to disable completely.
        epilog: Optional epilog shown in the help message.
        help_depth: How many subfields of a config to flatten
            for ``--help [PATH]``.

    Returns:
        The wrapped function.
    """

    def main_decorator(
        fn: Callable,
    ) -> Callable:
        wrapped = entry_point.EntryPoint(
            fn=fn,
            config=config,
            setup_logging=setup_logging,
            prog=prog,
            description=description,
            epilog=epilog,
            help_depth=help_depth,
        )
        return functools.wraps(fn)(wrapped)

    if fn is None:
        return main_decorator
    return main_decorator(fn)



from typing import (
    Any,
    ClassVar,
)

import marsh
from . import base


class MarshalSchemaMeta(base.SchemaMeta['MarshalSchema']):

    def __call__(
        cls,  # noqa: B902
        value,
        *args,
        **kwargs,
    ):
        if cls is not MarshalSchema:
            return type(object).__call__(cls, value, *args, **kwargs)
        return cls.registry.match(value).build(value, *args, **kwargs)


class MarshalSchema(base.Schema, metaclass=MarshalSchemaMeta):
    """Marshals its value input into a JSON-like object.

    Arguments:
        value: The value to match/marshal.
    """

    registry: ClassVar[base.SchemaRegistry['MarshalSchema']] = \
        base.SchemaRegistry(error=marsh.errors.MarshalError)

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        """Marshal the value held by this instance
        into a JSON-like object.

        Returns:
            The JSON-like marshaled value.
        """
        raise NotImplementedError


class WrapperMarshalSchema(
    base.WrapperSchema[MarshalSchema],
    MarshalSchema,
):

    def __init__(
        self,
        value: Any,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(value, *args, **kwargs)

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        return self.schema.marshal()



"""Path tools that work with any separator character."""
from typing import (
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Sequence,
)

from . import errors


def iterative_split(
    path: str,
    delimiter: str = '.',
) -> Iterator[str]:
    """Iterate through the fields in a path.

    Escaped or quoted (unescaped quotes) delimiters will become part of a field
    instead of splitting it into more fields. Quotes are removed
    from the final fields.

    Arguments:
        path: The path to split into fields and iterate through.
        delimiter: The delimiter between fields in the path.

    Returns:
        Field iterator.
    """
    if not path:
        return
    field = []
    quote: Optional[str] = None
    escaped: bool = False
    for char in path:
        if escaped:
            field.append(char)
            escaped = False
        elif char == '\\':
            escaped = True
        elif quote:
            if char == quote:
                quote = None
            else:
                field.append(char)
        elif char in ('"', "'"):
            quote = char
        elif char == delimiter:
            if field:
                yield ''.join(field)
            field = []
        else:
            field.append(char)
    if quote:
        raise errors.PathError(f'closing quote `{quote}` missing from path')
    if escaped:
        raise errors.PathError('escape character can not be last character of path')
    if field:
        yield ''.join(field)


def split(
    path: str,
    delimiter: str = '.',
) -> Sequence[str]:
    """Split the fields of a path.

    Escaped or quoted (unescaped quotes) delimiters will become part of a field
    instead of splitting it into more fields. Quotes are removed
    from the final fields.

    Arguments:
        path: The path to split into fields.
        delimiter: The delimiter between fields in the path.

    Returns:
        The fields.
    """
    return tuple(iterative_split(path=path, delimiter=delimiter))


def escape_field(
    field: str,
    delimiter: str = '.',
) -> str:
    """Escape all delimiters and quotes found in a field.

    This may be done to maintain the original field when it becomes
    part of a path that is later split back into fields.

    Arguments:
        field: The field to escape characters in.
        delimiter: The delimiter used.

    Returns:
        The escaped field.
    """
    chars = []
    for char in field:
        if char in ('"', "'", '\\', delimiter):
            chars.append('\\')
        chars.append(char)
    return ''.join(chars)


def strip_delimiter(
    field_or_path: str,
    delimiter: str = '.',
) -> str:
    """Remove any delimiters from the start and end of a field or path.

    Escaped delimiters are not removed.

    Arguments:
        field_or_path: The field or path to strip delimiters from.
        delimiter: The delimiter to strip.

    Returns:
        The field or path with stripped delimiters.
    """
    while field_or_path.startswith(delimiter):
        field_or_path = field_or_path[1:]
    while (
        field_or_path.endswith(delimiter)
        and not field_or_path.endswith(f'\\{delimiter}')
    ):
        field_or_path = field_or_path[:-1]
    return field_or_path


def prepend(
    path: str,
    field: str,
    delimiter: str = '.',
) -> str:
    """Prepend a field to a path.

    Arguments:
        field: The field to prepend to a path.
        path: The path to prepend the field to.
        delimiter: The delimiter used between fields in the path.

    Returns:
        The prepended path.
    """
    if not path:
        return field
    if not field:
        return path
    return delimiter.join(
        (
            escape_field(field, delimiter=delimiter),
            strip_delimiter(path, delimiter=delimiter),
        ),
    )


def append(
    path: str,
    field: str,
    delimiter: str = '.',
) -> str:
    """Append a field to a path.

    Arguments:
        field: The field to append to a path.
        path: The path to append the field to.
        delimiter: The delimiter used between fields in the path.

    Returns:
        The appended path.
    """
    if not path:
        return field
    if not field:
        return path
    return delimiter.join(
        (
            strip_delimiter(path, delimiter=delimiter),
            escape_field(field, delimiter=delimiter),
        ),
    )


def join_fields(
    fields: Iterable[str],
    delimiter: str = '.',
) -> str:
    """Create a path from fields.

    Arguments:
        fields: The fields to create a path from.
        delimiter: The delimiter to use in the path between fields.

    Returns:
        The path.
    """
    return delimiter.join(
        escape_field(field, delimiter)
        for field in filter(None, fields)
    )


class _head_return_value(NamedTuple):
    head: str
    path: str


def head(
    path: str,
    delimiter: str = '.',
) -> _head_return_value:
    """Split the first field in a path from the path.

    Arguments:
        path: The path to split.
        delimiter: The delimiter used between fields in the path.

    Returns:
        A tuple of the first field in the path and the remaining path.
    """
    fields = list(filter(None, split(path, delimiter=delimiter)))
    if not fields:
        fields = ['']
    return _head_return_value(
        head=fields[0],
        path=join_fields(fields[1:], delimiter=delimiter),
    )


class _tail_return_value(NamedTuple):
    tail: str
    path: str


def tail(
    path: str,
    delimiter: str = '.',
) -> _tail_return_value:
    """Split the last field in a path from the path.

    Arguments:
        path: The path to split.
        delimiter: The delimiter used between fields in the path.

    Returns:
        A tuple of the last field in the path and the remaining path.
    """
    fields = list(filter(None, split(path, delimiter=delimiter)))
    if not fields:
        fields = ['']
    return _tail_return_value(
        tail=fields[-1],
        path=join_fields(fields[:-1], delimiter=delimiter),
    )



"""Tools for inserting entry points into python applications."""
import argparse
import contextlib
import datetime
import logging
import os
import shutil
import sys
import textwrap
from typing import (
    Any,
    Callable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
)

import marsh
import marsh.doc
import marsh.parse


FIELD_INDENT = 2
DESCRIPTION_INDENT = 24
DESCRIPTION_MARGIN = 4


class EntryPoint:

    LOGGING_LEVELS = [
        'CRITICAL',
        'ERROR',
        'WARNING',
        'INFO',
        'DEBUG',
        'NOTSET',
    ]

    def __init__(
        self,
        fn: Callable,
        config: Optional[Any] = None,
        setup_logging: bool = False,
        prog: Optional[str] = None,
        description: Optional[str] = None,
        epilog: Optional[str] = None,
        help_depth: int = 2,
    ) -> None:
        self.fn = fn
        self.config = config
        self.setup_logging = setup_logging
        self.prog = prog
        self.description = description
        self.epilog = epilog
        self.help_depth = help_depth
        self.schema: marsh.schema.UnmarshalSchema
        if config is None or marsh.utils.is_missing(config):
            self.schema = marsh.schema.argument.ArgumentsUnmarshalSchema(fn)
        else:
            self.schema = marsh.schema.UnmarshalSchema(config)

    @property
    def terminal_width(
        self,
    ) -> int:
        return max(
            shutil.get_terminal_size().columns - 2,
            11,
        )

    def get_arg_parser(
        self,
    ) -> argparse.ArgumentParser:
        description = self.description
        if description is None:
            description = self.schema.doc_description()
        if description:
            description = textwrap.fill(
                description,
                width=self.terminal_width,
            )
        parser = argparse.ArgumentParser(
            prog=self.prog,
            description=description,
            epilog=self.epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            add_help=False,
        )

        parser.add_argument(
            '-h',
            '--help',
            dest='help',
            type=str,
            metavar='PATH',
            nargs='?',
            const="",
            help=(
                'Show this message and exit. Optionally display '
                'config documentation for a path in the config '
                'structure'
            ),
            default=None,
        )

        parser.add_argument(
            '-c',
            '--config-dir',
            dest='path',
            type=str,
            metavar='PATH',
            help=(
                'Root directory for config files. '
                'Config paths given in override '
                'arguments will be evaluated relative '
                '`--config-dir`. Defaults '
                'to the working directory of the caller.'
            ),
            default=None,
        )

        parser.add_argument(
            '-o',
            '--output',
            type=str,
            nargs='?',
            default=None,
            const='jobs/%Y%m%dT%H%M%S.%fZ',
            metavar='PATH',
            help=(
                'Change the current working directory '
                'when launching the application '
                '(after loading the final configuration). '
                'If no path is given when using this '
                'argument it defaults to "%(const)s"'
            ),
        )

        parser.add_argument(
            '--config-out',
            type=str,
            nargs='?',
            default=None,
            const='-',
            metavar='PATH',
            help=(
                'Write the current input configuration, '
                'then proceed. If no path is specified, the '
                'configuration is written to stdout. '
                'Relative paths are affected by the '
                '--output argument.'
            ),
        )

        if self.setup_logging:
            parser.add_argument(
                '--logging-level',
                type=str,
                metavar='LEVEL',
                help=(
                    'Specify the logging level. '
                    'Default: %(default)s.'
                ),
                choices=self.LOGGING_LEVELS,
                default='INFO',
            )

            parser.add_argument(
                '--logging-format',
                type=str,
                metavar='FORMAT',
                help=(
                    'Specify the logging format. '
                    'Default: %(default)s.'
                ),
                default='%(asctime)s %(levelname)s: %(message)s',
            )

        parser.add_argument(
            'overrides',
            help=(
                'Assign new values for fields in the '
                'configuration using `=` or assign the '
                'content of a config file using `@`. '
                'Adding `+` in front (`+=`, `+@`) combines '
                'the current value of the specified field '
                'with its new assigned value. '
                'Prepending a field with `~` removes it.'
            ),
            type=str,
            nargs='*',
        )
        return parser

    def load_config(
        self,
        args: argparse.Namespace,
    ) -> marsh.element.ElementType:
        root = None
        config = marsh.MISSING
        if args.path:
            root = args.path
        for override in map(marsh.parse.override, args.overrides):
            config = override.apply(
                element=config,
                root=root,
                keep_meta=True,
            )
        config = marsh.element.resolve(config)
        return marsh.config.drop_meta(config)

    @contextlib.contextmanager
    def maybe_change_cwd(
        self,
        args: argparse.Namespace,
    ) -> Iterator[None]:
        prev_path: Optional[str] = None
        if args.output:
            path = datetime.datetime.now(datetime.timezone.utc).strftime(args.output)
            prev_path = os.getcwd()

            def handle_create_dir_error(
                error: Exception,
            ) -> None:
                print(f'failed to create output directory {path}: {error}')
                sys.exit(1)

            with marsh.errors.maybe_handle_error(handle_create_dir_error):
                os.makedirs(path, exist_ok=True)
            os.chdir(path)
        yield
        if prev_path is not None:
            os.chdir(prev_path)

    def maybe_print_help(
        self,
        args: argparse.Namespace,
        passthrough_args: Optional[Sequence] = None,
        passthrough_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> None:
        if args.help is None:
            return
        if not args.help:
            self.get_arg_parser().print_help()
            if isinstance(
                self.schema,
                marsh.schema.template.StructuredUnmarshalSchema,
            ):
                doc = self.schema.doc(
                    depth=self.help_depth,
                    default_args=passthrough_args,
                    default_kwargs=passthrough_kwargs,
                )
            else:
                doc = self.schema.doc(depth=self.help_depth)
            field_help = marsh.doc.terminal.format_unmarshal_fields(
                source=doc,
                depth=self.help_depth,
            )
            if field_help:
                print('\nfields:')
                print(field_help)
            sys.exit(0)

        def handle_select_error(
            error: marsh.errors.MarshError,
        ) -> None:
            print(f'failed to traverse configuration: {error.pretty()}')
            sys.exit(1)

        with marsh.errors.maybe_handle_error(handle_select_error):
            schema = self.schema.select(args.help)

        help_msg = marsh.doc.terminal.format_unmarshal_help(
            source=schema,
            depth=self.help_depth,
        )
        if not help_msg:
            print(f'no help documention available for path: {args.help}')
            sys.exit(1)
        print(help_msg)
        sys.exit(0)

    def __call__(
        self,
        *args,
        **kwargs,
    ) -> Any:
        if args or kwargs:
            if not isinstance(
                self.schema,
                marsh.schema.argument.ArgumentsUnmarshalSchema,
            ):
                return self.fn(*args, **kwargs)
        cmd_args = self.get_arg_parser().parse_args()
        self.maybe_print_help(
            cmd_args,
            args,
            kwargs,
        )
        if self.setup_logging:
            logging.basicConfig(
                level=cmd_args.logging_level,
                format=cmd_args.logging_format,
            )

        def handle_load_config_error(
            error: marsh.errors.MarshError,
        ) -> None:
            print(f'failed to load configuration: {error.pretty(True)}')
            sys.exit(1)

        with marsh.errors.maybe_handle_error(handle_load_config_error):
            config = self.load_config(cmd_args)

        with self.maybe_change_cwd(cmd_args):
            if cmd_args.config_out:

                def handle_write_config_error(
                    error: Exception,
                ) -> None:
                    print('Failed to write input config to ', end='')
                    if cmd_args.config_out == '-':
                        print('stdout', end=': ')
                    else:
                        print(cmd_args.config_out, end=': ')
                    if isinstance(error, marsh.errors.MarshError):
                        print(error.pretty(True))
                    else:
                        print(error)
                    sys.exit(1)

                with marsh.errors.maybe_handle_error(
                    handle_write_config_error,
                    Exception,
                ):
                    marsh.config.write(config, path=cmd_args.config_out)

            def handle_unmarshal_error(
                error: marsh.errors.MarshError,
            ) -> None:
                print(f'failed to unmarshal config: {error.pretty(True)}')
                sys.exit(1)

            if isinstance(self.schema, marsh.schema.argument.ArgumentsUnmarshalSchema):
                with marsh.errors.maybe_handle_error(handle_unmarshal_error):
                    arguments = self.schema.unmarshal(
                        config,
                        default_args=args,
                        default_kwargs=kwargs,
                    )
                return self.fn(*arguments.args, **arguments.kwargs)
            with marsh.errors.maybe_handle_error(handle_unmarshal_error):
                value = self.schema.unmarshal(config)
            return self.fn(value)



import logging
from typing import (
    Any,
    Literal,
    Mapping,
    Union,
)

import marsh


Arg = Union[str, Mapping[Literal['name', 'level'], Union[int, str]]]


@marsh.schema.register
class LoggerMarshalSchema(marsh.schema.MarshalSchema):

    value: logging.Logger

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return isinstance(value, logging.Logger)
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~logging.Logger`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals a python logger into a mapping '
            'with keys ``"name"`` and ``"level"``.'
        )

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        element: dict = dict(name=self.value.name)
        level = logging.getLevelName(self.value.level)
        if level == 'NOTSET':
            return element
        element.update(level=level)
        try:
            if level.startswith('Level '):
                element.update(level=int(level.split(' ')[-1]))
        except Exception:
            pass
        return element


@marsh.schema.register
class LoggerUnmarshalSchema(marsh.schema.UnmarshalSchema[logging.Logger]):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return value is logging.Logger

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~logging.Logger`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'A python logging.Logger. If no '
            'arguments are given the root '
            'logger is used. Takes either '
            'a single string which is used as name '
            'for the logger or a mapping with optional '
            'keys "name" and "level". The level can be '
            'given as a string or an integer.'
        )

    def doc_field_type(
        self,
    ) -> str:
        return 'Logger(...)'

    def doc_description(
        self,
    ) -> str:
        return self.doc_static_description()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> logging.Logger:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            return logging.getLogger()
        try:
            arg: Arg = marsh.unmarshal(Arg, element)  # type: ignore
        except Exception:
            raise marsh.errors.UnmarshalError(
                (
                    'failed to unmarshal input element into a '
                    'string or mapping with optional keys `name` '
                    'and `level`'
                ),
                element=element,
                type=self.value,
            )
        if isinstance(arg, str):
            return logging.getLogger(arg)
        if 'name' in arg:
            logger = logging.getLogger(str(arg['name']))
        else:
            logger = logging.getLogger()
        if 'level' in arg:
            try:
                logger.setLevel(arg['level'])
            except Exception:
                raise marsh.errors.UnmarshalError(
                    f'could not set logging level to "{arg["level"]}"',
                    element=element,
                    type=type,
                )
        return logger



import functools
import types
from typing import (
    Any,
    Sequence,
    TypeVar,
    get_type_hints,
)

import marsh
from . import mapping


_T = TypeVar('_T')


class _Sentinel:

    def __str__(
        self,
    ) -> str:
        return ''


@marsh.schema.register(lower_priority=mapping.MappingUnmarshalSchema)
class TypedDictUnmarshalSchema(marsh.schema.template.StructuredUnmarshalSchema[_T]):

    @functools.cached_property
    def required_keys(
        self,
    ) -> Sequence[str]:
        if hasattr(self.value, '__required_keys__'):
            keys = self.value.__required_keys__  # type: ignore
        elif self.value.__total__:  # type: ignore
            keys = frozenset(get_type_hints(self.value))
        else:
            keys = frozenset()
        ordered_keys = tuple(get_type_hints(self.value))
        return sorted(keys, key=lambda x: ordered_keys.index(x))

    @functools.cached_property
    def optional_keys(
        self,
    ) -> Sequence[str]:
        if hasattr(self.value, '__optional_keys__'):
            keys = self.value.__optional_keys__  # type: ignore
        elif not self.value.__total__:  # type: ignore
            keys = frozenset(get_type_hints(self.value))
        else:
            keys = frozenset()
        ordered_keys = tuple(get_type_hints(self.value))
        return sorted(keys, key=lambda x: ordered_keys.index(x))

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._sentinel = _Sentinel()
        annotations = get_type_hints(self.value)
        self.schemas = types.MappingProxyType(
            {
                **{
                    name: marsh.schema.UnmarshalSchema(
                        annotations[name],
                    )
                    for name in self.required_keys
                },
                **{
                    name: marsh.schema.UnmarshalSchema(
                        annotations[name],
                        default=self._sentinel,
                    )
                    for name in self.optional_keys
                },
            },
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_typeddict_type(value)

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~typing.TypedDict`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals any subclass of :class:`~typing.TypedDict` '
            'from a mapping input. If ``total=False`` then '
            'all keys are optional.'
        )

    def construct(
        self,
        *args,
        **kwargs,
    ) -> _T:
        if args and not kwargs:
            kwargs = {
                key: value
                for key, value
                in zip(get_type_hints(self.value), args)
            }
            args = ()
        mapping = dict(*args, **kwargs)
        for key, value in tuple(mapping.items()):
            if value is self._sentinel:
                if key in self.optional_keys:
                    del mapping[key]
                else:
                    raise marsh.errors.MissingValueError(
                        f'required key: {key}',
                    )
        return mapping  # type: ignore



"""This module creates support for inheritance-based type matching.

For all types in a namespace each subclass of a type being
that is being unmarshaled is considered."""
import collections.abc
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    ItemsView,
    Iterator,
    KeysView,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    TypedDict,
    Union,
    ValuesView,
    cast,
    overload,
)

import marsh
from . import core
from .core.base import caches


_T = TypeVar('_T')
_K = TypeVar('_K')
_V = TypeVar('_V')


class _SingletonMapping(
    Generic[_K, _V],
    metaclass=marsh.utils.SingletonMeta,
):

    def __getitem__(
        self,
        key: _K,
    ) -> _V:
        raise NotImplementedError

    def __len__(
        self,
    ) -> int:
        raise NotImplementedError

    def __iter__(
        self,
    ) -> Iterator[_K]:
        raise NotImplementedError

    def get(
        self,
        key,
        default: Any = None,
    ) -> Union[_V, Any]:
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def keys(
        self,
    ) -> KeysView[_K]:
        return collections.abc.KeysView(self)  # type: ignore

    def items(
        self,
    ) -> ItemsView[_K, _V]:
        return collections.abc.ItemsView(self)  # type: ignore

    def values(
        self,
    ) -> ValuesView[_V]:
        return collections.abc.ValuesView(self)  # type: ignore

    def __eq__(
        self,
        other: Any,
    ) -> bool:
        if not isinstance(other, _SingletonMapping):
            return NotImplemented
        return dict(self.items()) == dict(other.items())


class Namespaces(_SingletonMapping[str, 'Namespace']):
    """Singleton class holding all namespaces.

    Includes functionality for matching a type
    against the classes of all namespaces. These
    matching functions are cached for improved
    performance."""

    class CacheInfo(TypedDict):
        find_class: marsh.utils.CacheInfo
        find_subclasses: marsh.utils.CacheInfo
        find_namespaces: marsh.utils.CacheInfo

    class FullCacheInfo(CacheInfo):
        names: Mapping[str, 'Namespace.CacheInfo']

    def __init__(
        self,
    ) -> None:
        self._namespaces: Dict[str, 'Namespace'] = {}

    def __getitem__(
        self,
        name: str,
    ) -> 'Namespace':
        if name not in self:
            raise KeyError(
                marsh.utils.get_closest_error_message(
                    value=name,
                    candidates=self,
                    key='namespace',
                ),
            )
        return self._namespaces[name]

    def __delitem__(
        self,
        name,
    ) -> None:
        del self._namespaces[name]

    def __contains__(
        self,
        name,
    ) -> bool:
        return name in self._namespaces

    def __len__(
        self,
    ) -> int:
        return len(self._namespaces)

    def __iter__(
        self,
    ) -> Iterator[str]:
        yield from self._namespaces

    def find_subclasses_iter(
        self,
        cls: Any,
    ) -> Iterator[Tuple[str, 'Namespace']]:
        """Yield results from finding subclasses of an input type
        in all namespaces.

        This method provides results for the same but cached
        method :func:`~Namespaces.find_subclasses`.

        Arguments:
            cls: The input type.

        Returns:
            Iterator of result.
        """
        for namespace in self.values():
            for name in namespace.find_subclasses(cls):
                yield name, namespace

    def find_namespaces_iter(
        self,
        cls: Any,
    ) -> Iterator['Namespace']:
        """Yield results from finding namespaces matching an input type.

        This method provides results for the same but cached
        method :func:`~Namespaces.find_namespaces`.

        Arguments:
            cls: The input type.

        Returns:
            Iterator of result.
        """
        for namespace in self._namespaces.values():
            try:
                if issubclass(cls, namespace.base):
                    yield namespace
            except Exception:
                pass

    def new(
        self,
        name: str,
        base: Type[_T],
    ) -> 'Namespace[_T]':
        """Create a new namespace.

        The created namespace has functions for registering
        new items and building them.

        Arguments:
            name: Name of namespace.
            base: Base class for namespace.

        Returns:
            The namespace.
        """
        if name in self:
            raise marsh.errors.MarshError(
                f'namespace "{name}" already exists',
            )
        self._namespaces[name] = (
            namespace := Namespace(
                name=name,
                base=base,
            )
        )
        return namespace

    def cache_clear(
        self,
        full: bool = False,
    ) -> None:
        """Clear the cache of this class.

        Arguments:
            full: If ``True``, also clear the cache
                of all namespaces.
        """
        self.find_class.cache_clear()  # type: ignore
        self.find_subclasses.cache_clear()  # type: ignore
        self.find_namespaces.cache_clear()  # type: ignore
        if full:
            for namespace in self._namespaces.values():
                namespace.cache_clear()

    @overload
    def cache_info(
        self,
        full: Literal[True],
    ) -> 'Namespaces.FullCacheInfo':
        ...

    @overload
    def cache_info(
        self,
    ) -> 'Namespaces.CacheInfo':
        ...

    def cache_info(
        self,
        full: bool = False,
    ) -> Union['Namespaces.CacheInfo', 'Namespaces.FullCacheInfo']:
        """Get info for the cache of this class.

        Arguments:
            full: If ``True``, return the cache info
                for all namespaces as well.

        Returns:
            The cache info.
        """
        info: dict = dict(
            find_class=marsh.utils.CacheInfo.from_info(
                self.find_class.cache_info(),  # type: ignore
            ),
            find_subclasses=marsh.utils.CacheInfo.from_info(
                self.find_subclasses.cache_info(),  # type: ignore
            ),
            find_namespaces=marsh.utils.CacheInfo.from_info(
                self.find_namespaces.cache_info(),  # type: ignore
            ),
        )
        if full:
            names = {}
            for name, namespace in self.items():
                names[name] = namespace.cache_info()
            info['names'] = names
        return info  # type: ignore

    @caches.new_callable_cache(
        name='marsh.namespaces.find_class',
        safe=True,
        binding='weak',
    )
    def find_class(
        self,
        cls: Any,
    ) -> Optional[str]:
        """Search through all namespaces and return
        the name of the first class that equals the given type.

        Arguments:
            cls: The type to find a name for.

        Returns:
            A name if found, else :data:`None`.
        """
        for namespace in reversed(self._namespaces.values()):
            if (name := namespace.find_class(cls)):
                return name
        return None

    @caches.new_callable_cache(
        name='marsh.namespaces.find_subclasses',
        safe=True,
        binding='weak',
    )
    def find_subclasses(
        self,
        cls: Any,
    ) -> marsh.utils.IterableFromIterator[Tuple[str, 'Namespace']]:
        """Search for all subclasses of the given type.

        Argument:
            cls: The type to find subclasses for.

        Returns:
            Iterable of names and namespaces.
        """
        return marsh.utils.IterableFromIterator(self.find_subclasses_iter(cls))

    @caches.new_callable_cache(
        name='marsh.namespaces.find_namespaces',
        safe=True,
        binding='weak',
    )
    def find_namespaces(
        self,
        cls: Any,
    ) -> marsh.utils.IterableFromIterator['Namespace']:
        """Find the namespaces with base classes that
        are superclasses of the given type.

        Arguments:
            cls: The type to find namespaces for.

        Returns:
            Iterable of namespaces found.
        """
        return marsh.utils.IterableFromIterator(self.find_namespaces_iter(cls))


class Namespace(Mapping[str, core.unmarshal.UnmarshalSchema[_T]]):
    """Holds a group of subclasses for a base class which are each
    associated with a name.

    This class should be initialized via :func:`Namespaces.new`.

    Includes functionality for matching a type against
    the classes held by this namespace. This functionality
    is cached for improved performance.

    Arguments:
        base: The base class for the namespace.
    """

    class CacheInfo(TypedDict):
        find_class: marsh.utils.CacheInfo
        find_subclasses: marsh.utils.CacheInfo

    components: Dict[str, Type[_T]]
    _namespaces: Namespaces = Namespaces()

    def __init__(
        self,
        name: str,
        base: Type[_T],
    ) -> None:
        self.name = name
        self.base = base
        self.components = {}
        self._schemas: Dict[str, core.unmarshal.UnmarshalSchema[_T]] = {}
        # when wrapping functions on an instance the `self`
        # argument is not included until the most inner
        # wrapped function is called which means that we
        # skip setting `binding='ignore'`
        self.find_class = caches.new_callable_cache(  # type: ignore
            self.find_class,
            name=f'marsh.namespaces.{name}.find_class',
            safe=True,
        )
        self.find_subclasses = caches.new_callable_cache(  # type: ignore
            self.find_subclasses,
            name=f'marsh.namespaces.{name}.find_subclasses',
            safe=True,
        )

    def __getitem__(
        self,
        name: str,
    ) -> core.unmarshal.UnmarshalSchema[_T]:
        if name not in self:
            raise KeyError(
                marsh.utils.get_closest_error_message(
                    value=name,
                    candidates=self,
                    key='name',
                ),
            )
        if name not in self._schemas:
            # instead of picking out the matched
            # schema types we need to run the normal
            # constructor so that the type is cached.
            # If not, the recursive tests for namespace
            # components fail.
            schema = marsh.schema.UnmarshalSchema[_T](
                self.components[name],
            )
            # extract inner schema
            while isinstance(
                schema,
                marsh.schema.types.namespace.NamespaceUnmarshalSchema,
            ):
                schema = schema.schema
            self._schemas[name] = schema
        return self._schemas[name]

    def __iter__(
        self,
    ) -> Iterator[str]:
        yield from self.components

    def __len__(
        self,
    ) -> int:
        return len(self.components)

    def __contains__(
        self,
        name: object,
    ) -> bool:
        return name in self.components

    def cache_clear(
        self,
    ) -> None:
        """Clear the cache."""
        self.find_class.cache_clear()  # type: ignore
        self.find_subclasses.cache_clear()  # type: ignore

    def cache_info(
        self,
    ) -> 'Namespace.CacheInfo':
        """Get info for the cache.

        Returns:
            The cache info.
        """
        return dict(
            find_class=self.find_class.cache_info(),  # type: ignore
            find_subclasses=self.find_subclasses.cache_info(),  # type: ignore
        )

    def find_class(
        self,
        cls: Type[_T],
    ) -> Optional[str]:
        """Find the name of a class maybe held by this namespace.

        Arguments:
            cls: The type to find a name for.

        Returns:
            :data:`None` if no class was matched, else the name
            associated with the type
        """
        if not issubclass(cls, self.base):
            return None
        for name, component in self.components.items():
            if component == cls:
                return name
        return None

    def find_subclasses_iter(
        self,
        cls: Type[_T],
    ) -> Iterator[str]:
        """Yield names of components in this namespace that are
        subclasses of the input type.

        This method provides results for the same but cached
        method :func:`~Namespace.find_subclasses`.

        Arguments:
            cls: The input type.

        Returns:
            Iterator of string names.
        """
        for name, component in self.components.items():
            try:
                if issubclass(component, cls):
                    yield name
            except Exception:
                pass

    def find_subclasses(
        self,
        cls: Type[_T],
    ) -> marsh.utils.IterableFromIterator[str]:
        """Get the names of all classes held by this namespace
        that are subclasses of the input.

        Arguments:
            cls: The type to find subclasses for.

        Returns:
            The names of the subclasses.
        """
        return marsh.utils.IterableFromIterator(self.find_subclasses_iter(cls))

    @overload
    def register(
        self,
        component: Type[_T],
        *,
        name: str,
        replace: bool = False,
    ) -> Type[_T]:
        ...

    @overload
    def register(
        self,
        *,
        name: str,
        replace: bool = False,
    ) -> Callable[[Type[_T]], Type[_T]]:
        ...

    def register(
        self,
        component: Optional[Type[_T]] = None,
        *,
        name: str,
        replace: bool = False,
    ) -> Union[Callable[[Type[_T]], Type[_T]], Type[_T]]:
        """Register a new type for this namespace.

        Arguments:
            name: The name to associate with the type.
            replace: Allow replacing any existing value
                with the same name.

        Returns:
            A decorator for registering the type if no
            type was given else the type itself.
        """
        if name in self:
            if not replace:
                raise marsh.errors.MarshError(
                    f'a component with the name "{name}" '
                    'already exists',
                )
            self._schemas.pop(name, None)

        def _register(
            component: Type[_T],
        ) -> Type[_T]:
            if not issubclass(component, self.base):
                raise ValueError(
                    'registered component must be '
                    F'a subclass of {self.base}: {component}',
                )
            self.components[name] = component
            self.cache_clear()
            self._namespaces.cache_clear()
            marsh.schema.UnmarshalSchema.cache_clear()
            return component

        if component is None:
            return _register
        return _register(component)

    @overload
    def build(
        self,
        element: marsh.element.ElementType = marsh.MISSING,
        *,
        name: str,
    ) -> _T:
        ...

    @overload
    def build(
        self,
        element: Optional[marsh.element.SequenceElementType] = marsh.MISSING,
        *args,
        name: str,
    ) -> _T:
        ...

    @overload
    def build(
        self,
        element: Optional[marsh.element.MappingElementType] = marsh.MISSING,
        *,
        name: str,
        **kwargs,
    ) -> _T:
        ...

    def build(
        self,
        element: marsh.element.ElementType = marsh.MISSING,
        *args,
        name: str,
        **kwargs,
    ) -> _T:
        """Unmarshal a type held by this namespace."""
        if name not in self:
            raise marsh.errors.UnmarshalError(
                marsh.utils.get_closest_error_message(
                    value=name,
                    candidates=self,
                    key='name',
                ),
                element=element,
            )
        if args and kwargs:
            raise TypeError(
                'variable positional arguments can not be used '
                'simultaneously as variable keyword arguments',
            )
        if args:
            if marsh.utils.is_sequence(element):
                element = list(element or ())
                element.extend(cast(Sequence, marsh.marshal(args)))
            elif element is None or marsh.utils.is_missing(element):
                element = marsh.marshal(args)
            else:
                raise TypeError(
                    'variable positional arguments can only be used '
                    'when `element` is a sequence.',
                )
        if kwargs:
            if marsh.utils.is_mapping(element):
                element = dict(element)
                element.update(cast(Mapping, marsh.marshal(kwargs)))
            elif element is None or marsh.utils.is_missing(element):
                element = marsh.marshal(kwargs)
            else:
                raise TypeError(
                    'variable keyword arguments can only be used '
                    'when `element` is a mapping.',
                )
        return self[name].unmarshal(element)



import re
import types
from typing import (
    Any,
    Optional,
    TypeVar,
    get_type_hints,
)

import marsh
from . import (
    sequence,
    tuple,
)


_T = TypeVar('_T', bound=marsh.utils.NamedTupleProtocol)


def is_generated_class_docstring(
    docstring: Optional[str],
) -> bool:
    if docstring:
        return bool(re.match(r'^[a-zA-Z0-9_]+\(([a-zA-Z0-9_]+,)*\)$', docstring))
    return False


@marsh.schema.register
class NamedTupleMarshalSchema(marsh.schema.MarshalSchema):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_namedtuple(value)

    @staticmethod
    def doc_static_type() -> str:
        return ':func:`~collections.namedtuple`, :class:`~typing.NamedTuple`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals namedtuples into a mapping '
            'where the keys correspond to the field '
            'names of the namedtuple. Supports both '
            ':func:`~collections.namedtuple` and :class:`~typing.NamedTuple`.'
        )

    def marshal(
        self,
    ) -> dict:
        return {
            name: marsh.marshal(value)
            for name, value
            in zip(self.value._fields, self.value)
        }


@marsh.schema.register(
    lower_priority=(
        sequence.SequenceUnmarshalSchema,
        tuple.TupleUnmarshalSchema,
    ),
)
class NamedTupleUnmarshalSchema(marsh.schema.template.StructuredUnmarshalSchema[_T]):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        schemas = {}
        annotations = get_type_hints(self.value)
        for field in self.value._fields:
            with marsh.errors.prepend(field):
                schemas[field] = marsh.schema.UnmarshalSchema(
                    annotations.get(field, Any),
                    default=self.value._field_defaults.get(field, marsh.MISSING),
                )
        self.schemas = types.MappingProxyType(schemas)

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_namedtuple_type(value)

    @staticmethod
    def doc_static_type() -> str:
        return ':func:`~collections.namedtuple`, :class:`~typing.NamedTuple`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals namedtuple types from a mapping '
            'or sequence input. Supports both '
            ':func:`~collections.namedtuple` and :class:`~typing.NamedTuple`.'
        )

    def doc_description(
        self,
    ) -> Optional[str]:
        doc = getattr(self.value, '__doc__', None)
        if not doc or is_generated_class_docstring(doc):
            return None
        return doc



from typing import (
    Any,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

import marsh


_T = TypeVar('_T')


@marsh.schema.register
class UnionUnmarshalSchema(marsh.schema.template.UnionUnmarshalSchema[_T]):
    """Attempts to unmarshal the types in a union one at a time in the order
    of the types held by the Union type alias."""

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.schemas = tuple(
            map(
                marsh.schema.UnmarshalSchema,
                get_args(self.value),
            ),
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return bool(
            get_origin(value) is Union
            and get_args(value),
        )

    @staticmethod
    def doc_static_type() -> str:
        return ':data:`~typing.Union`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'The union type which in later versions of '
            'python may be specified using the ``|`` operator. '
            'The order of the types in the union matter as an '
            'attempt is made to unmarshal each type where the value '
            'of the first successful unmarshal is returned. :data:`~typing.Optional` '
            'will always set :data:`None` as the second type in the produced union.'
        )
